from collections import OrderedDict

from django.core.exceptions import PermissionDenied, ImproperlyConfigured

import graphene
from graphene.types.mutation import MutationOptions
from graphene.types.utils import yank_fields_from_attrs

from graphene_django.forms.converter import convert_form_field


### Generic helpers

def get_user_from_info(info, allow_anonymous=False):
    """
    Convenience method for getting the user object from the info argument
    that a Graphene Mutation receives in the `mutate` method.

    This exists because otherwise the same four lines would have to repeated
    in every mutation where a user is required.

    :param info: The info object as received by a `mutate` method on
      a graphene.Mutation.
    :param allow_anonymous: If True, anonymous user instances are returned;
      if False, they are replaced with None.
    """
    user = getattr(info.context, 'user', None)
    if user is not None and (user.is_anonymous and not allow_anonymous):
        user = None
    return user


def make_failable_objecttype(orig_type, name=None):
    meta = getattr(orig_type, '_meta', None)
    if meta is None and name is None:
        raise ImproperlyConfigured(
            "The original type passed doesn't have a name that we can generate "
            "a name from, please specify one explicitly."
        )
    new_name = name if name else 'Failable{}'.format(meta.name)

    class FailableWrapper(graphene.ObjectType):
        error = graphene.String()
        success = graphene.NonNull(graphene.Boolean)
        result = graphene.Field(orig_type)

        class Meta:
            name = new_name
    return graphene.Field(FailableWrapper)


### Mutations

class MutationExecutionException(Exception):
    pass


class ValidationError(graphene.ObjectType):
    field = graphene.String()
    messages = graphene.List(graphene.String)


class ValidationErrors(graphene.ObjectType):
    validation_errors = graphene.List(ValidationError)


class ExecutionError(graphene.ObjectType):
    error_message = graphene.NonNull(graphene.String)


class MutationError(graphene.Union):
    class Meta:
        types = (
            ValidationErrors,
            ExecutionError
        )


class FailableMutation(graphene.Mutation):
    class Meta:
        abstract = True

    error = MutationError()
    success = graphene.NonNull(graphene.Boolean)



### Mutations from Forms

def fields_for_form(form, only_fields, exclude_fields):
    fields = OrderedDict()
    for name, field in form.fields.items():
        is_not_in_only = only_fields and name not in only_fields
        is_excluded = name in exclude_fields
        if is_not_in_only or is_excluded:
            continue

        fields[name] = convert_form_field(field)
    return fields


class FormMutationOptions(MutationOptions):
    form_class = None
    fields_from_cleaned = []
    permissions = None


class FormMutation(FailableMutation):
    class Meta:
        abstract = True

    @classmethod
    def mutate(cls, root, info, **input):
        form = cls.get_form(root, info, **input)

        # Check form.is_valid() to see if we should mutate,
        # or return with ValidationErrors right away.
        if form.is_valid():
            try:
                result = cls.perform_mutate(form, info)
                # Do some defaulting for the `success` field based on the errors
                if not hasattr(result, 'success'):
                    result.success = getattr(result, 'error') is None
                return result
            except MutationExecutionException as e:
                return cls(
                    error=ExecutionError(error_message=str(e)),
                    success=False
                )
        else:
            validation_errors = [
                ValidationError(field=key, messages=value)
                for key, value in form.errors.items()
            ]
            return cls(
                error=ValidationErrors(validation_errors=validation_errors),
                success=False
            )

    @classmethod
    def get_form(cls, root, info, **input):
        form_kwargs = cls.get_form_kwargs(root, info, **input)
        return cls._meta.form_class(**form_kwargs)

    @classmethod
    def get_form_kwargs(cls, root, info, **input):
        kwargs = {"data": input}

        pk = input.pop("id", None)
        if pk:
            instance = cls._meta.model._default_manager.get(pk=pk)
            kwargs["instance"] = instance

        return kwargs

    @classmethod
    def __init_subclass_with_meta__(
        cls, form_class=None, only_fields=(), exclude_fields=(),
        _meta=None, **options
    ):
        if not form_class:
            raise Exception("form_class is required for FormMutation")

        form = form_class()
        arguments = yank_fields_from_attrs(
            fields_for_form(form, only_fields, exclude_fields),
            _as=graphene.Argument
        )
        if not _meta:
            _meta = FormMutationOptions(cls)
        _meta.form_class = form_class

        super().__init_subclass_with_meta__(
            _meta=_meta,
            arguments=arguments,
            **options
        )

    @classmethod
    def perform_mutate(cls, form, info):
        raise NotImplementedError(
            "perform_mutate needs to be overridden in the subclass!"
        )



import json

from django.utils.module_loading import import_string
from django.test import RequestFactory, TestCase

import graphene.test

class GrapheneAPITest(TestCase):
    base = '/graphql/'
    schema_src = 'some.schema'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.schema = import_string(self.schema_src)

    # adapted from https://stackoverflow.com/a/47762174/3090225
    def execute(self, query, user=None, variables=None, files=None,  **kwargs):
        # Construct context
        request_factory = RequestFactory()
        context = request_factory.get(self.base)
        context.user = user
        if files and type(files) == dict:
            for file in files:
                context.FILES[file] = files[file]

        # Create Graphene test client instance and use it to execute the query
        client = graphene.test.Client(self.schema)
        result = client.execute(
            query,
            context=context,
            variables=variables,
            **kwargs
        )
        return result

    def assertExecuteReturns(self, query, expected_result, **kwargs):
        """
        Custom assert function to idiomatically test the common pattern of
        executing a query and comparing the result dict to an expected
        result dict.

        :param query: The GraphQL query to execute.
        :param expected_result: The expected result dict to compare against.
        """
        return self.assertEqual(
            self.execute(query, **kwargs),
            expected_result
        )

class GrapheneMutationTest(GrapheneAPITest):
    mutation = None

    def mutate(self, user=None, variables=None, **kwargs):
        return self.execute(self.mutation, user, variables, **kwargs)


class GrapheneFormMutationTest(GrapheneMutationTest):
    gql_mutation_name = None
    
    def get_data(self, result):
        if not isinstance(result, dict):
            raise ValueError("Must pass a dict as value to get_data()!")
        if not 'data' in result or\
           not self.gql_mutation_name in result['data'] or\
           result['data'][self.gql_mutation_name] is None:
            raise Exception("Could not access data on result object {}!".format(result))
        return result['data'][self.gql_mutation_name]

    def get_error(self, result):
        err = self.get_data(result)['error'] or {}
        validation_errs = err.get('validationErrors')
        execution_err = err.get('errorMessage')

        if validation_errs is not None:
            return validation_errs
        elif execution_err is not None:
            return execution_err
        return None

    def assertSuccessful(self, result):
        """
        Asserts that the result is the result of a successfully executed
        mutation. This means that:
          - result['data'][name]['success'] is present and True, and
          - result['data'][name]['error'] is None
        (where `name` is self.gql_mutation_name).
        """
        self.assertIn('data', result)
        self.assertIn(self.gql_mutation_name, result['data'])
        self.assertTrue(
            self.get_data(result).get('success', False),
            str(self.get_data(result))
        )
        self.assertIsNone(self.get_error(result))

    def assertThrew(self, result):
        """
        Asserts that the result is the result of a mutation that was executed
        and threw an internal exception that was not caught. This means that:
          - result['data'] is None
          - result['errors'] is present
        """
        self.assertIsNone(result['data'])
        self.assertIsNotNone(result['errors'])

    def assertErrored(self, result):
        """
        Asserts that the result is the result of a non-successfully executed
        mutation. This means that:
          - result['data'][name]['success'] is present and False, and
          - result['data'][name]['error'] is present
        (where `name` is self.gql_mutation_name).
        """
        self.assertIn('data', result)
        self.assertIn(self.gql_mutation_name, result['data'])
        self.assertFalse(
            self.get_data(result).get('success', True),
            str(self.get_data(result))
        )
        self.assertIsNotNone(self.get_error(result))
